/******************************************************************************
 *  Library to store and retrieve data from EEPROM or write to the
 *  DAC on MX4cK
 *  Based on code from Microchip, see documentation below.
 *  Change configurations by changing defines in I2C_EEPROM.h
 * 
 * Mason, MEGR304 Seattle U, 2013
 * ***************************************************************************

 * ****************************************************************************
 *
 *  FileName:        I2C_master.c
 * Dependencies:
 * Processor:       PIC32
 *
 * Complier:        MPLAB C32
 *                  MPLAB IDE
 * Company:         Microchip Technology Inc.
 *
 * Software License Agreement
 *
 * The software supplied herewith by Microchip Technology Incorporated
 * (the ?Company?) for its PIC32 Microcontroller is intended
 * and supplied to you, the Company?s customer, for use solely and
 * exclusively on Microchip PIC32 Microcontroller products.
 * The software is owned by the Company and/or its supplier, and is
 * protected under applicable copyright laws. All rights are reserved.
 * Any use in violation of the foregoing restrictions may subject the
 * user to criminal sanctions under applicable laws, as well as to
 * civil liability for the breach of the terms and conditions of this
 * license.
 *
 * THIS SOFTWARE IS PROVIDED IN AN ?AS IS? CONDITION. NO WARRANTIES,
 * WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED
 * TO, IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE COMPANY SHALL NOT,
 * IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
 * CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
 *
 *********************************************************************
 * File Description: I2C Peripheral Lib Demo
 *
 * Change History:
 * Name        Date            Changes
 * A Rafiq     2/7/07          Initial Version
 * JM          2/7/09          Updated
 *
 * $Id: i2c_master.c 9558 2008-07-04 10:08:49Z deshmukha $
 *
 * Demo Code Description:
 * This demo code excersize the i2c channel 1 by interfacing
 * to 24LC256 serial eeprom.  The code writes to a location
 * and then verify the contents of the eeprom.  I2C Channel 1
 * is hooked up to 24LC256.  Address lines of the eeprom are
 * all tied to Vss.
 ********************************************************************/
#include <plib.h>
#include "I2C_DACEEPROM.h"



// ****************************************************************************
// ****************************************************************************
// Local Support Routines
// ****************************************************************************
// ****************************************************************************

/*******************************************************************************
  Function:
    BOOL StartTransfer( BOOL restart )

  Summary:
    Starts (or restarts) a transfer to/from the EEPROM.

  Description:
    This routine starts (or restarts) a transfer to/from the EEPROM, waiting (in
    a blocking loop) until the start (or re-start) condition has completed.

  Precondition:
    The I2C module must have been initialized.

  Parameters:
    restart - If FALSE, send a "Start" condition
            - If TRUE, send a "Restart" condition

  Returns:
    TRUE    - If successful
    FALSE   - If a collision occured during Start signaling

  Example:
    <code>
    StartTransfer(FALSE);
    </code>

  Remarks:
    This is a blocking routine that waits for the bus to be idle and the Start
    (or Restart) signal to complete.
 *****************************************************************************/

BOOL I2C_StartTransfer(BOOL restart) {
    I2C_STATUS status;

    // Send the Start (or Restart) signal
    if (restart) {
        I2CRepeatStart(EEPROM_I2C_BUS);
    } else {
        // Wait for the bus to be idle, then start the transfer
        while (!I2CBusIsIdle(EEPROM_I2C_BUS));

        if (I2CStart(EEPROM_I2C_BUS) != I2C_SUCCESS) {
            //DBPRINTF("Error: Bus collision during transfer Start\n");
            return FALSE;
        }
    }

    // Wait for the signal to complete
    do {
        status = I2CGetStatus(EEPROM_I2C_BUS);

    } while (!(status & I2C_START));

    return TRUE;
}

/*******************************************************************************
  Function:
    BOOL TransmitOneByte( UINT8 data )

  Summary:
    This transmits one byte to the EEPROM.

  Description:
    This transmits one byte to the EEPROM, and reports errors for any bus
    collisions.

  Precondition:
    The transfer must have been previously started.

  Parameters:
    data    - Data byte to transmit

  Returns:
    TRUE    - Data was sent successfully
    FALSE   - A bus collision occured

  Example:
    <code>
    TransmitOneByte(0xAA);
    </code>

  Remarks:
    This is a blocking routine that waits for the transmission to complete.
 *****************************************************************************/

BOOL I2C_TransmitOneByte(UINT8 data) {
    // Wait for the transmitter to be ready
    while (!I2CTransmitterIsReady(EEPROM_I2C_BUS));

    // Transmit the byte
    if (I2CSendByte(EEPROM_I2C_BUS, data) == I2C_MASTER_BUS_COLLISION) {
        return FALSE; // "Error: I2C Master Bus Collision\n"
    }

    // Wait for the transmission to finish
    while (!I2CTransmissionHasCompleted(EEPROM_I2C_BUS));

    return TRUE;
}

/*******************************************************************************
  Function:
    void StopTransfer( void )

  Summary:
    Stops a transfer to/from the EEPROM.

  Description:
    This routine Stops a transfer to/from the EEPROM, waiting (in a
    blocking loop) until the Stop condition has completed.

  Precondition:
    The I2C module must have been initialized & a transfer started.

  Parameters:
    None.

  Returns:
    None.

  Example:
    <code>
    StopTransfer();
    </code>

  Remarks:
    This is a blocking routine that waits for the Stop signal to complete.
 *****************************************************************************/

void I2C_StopTransfer(void) {
    I2C_STATUS status;

    // Send the Stop signal
    I2CStop(EEPROM_I2C_BUS);

    // Wait for the signal to complete
    do {
        status = I2CGetStatus(EEPROM_I2C_BUS);
    } while (!(status & I2C_STOP));
}

/* ----------------------
 * Initialize the I2C
 *  must call before using other functions
 * --------------------------*/
BOOL I2C_Initialize() {
    // Set the I2C baudrate
    UINT32 actualClock;
    actualClock = I2CSetFrequency(EEPROM_I2C_BUS, GetPeripheralClock(), I2C_CLOCK_FREQ);
    if (abs(actualClock - I2C_CLOCK_FREQ) > I2C_CLOCK_FREQ / 10) {
        return FALSE;
        // DBPRINTF("Error: I2C1 clock frequency (%u) error exceeds 10%%.\n", (unsigned) actualClock);
    }

    // Enable the I2C bus
    I2CEnable(EEPROM_I2C_BUS, TRUE);
}

/* --------------------------------------
 *
 *  Store array of data on EEPROM starting at specific address
 *
 *  Parameters:
 *     data = array of UNIT8 data
 *     startaddr =  address in eeprom to write data
 *     ndata = number of elements to write
 *
 *  Return: true if successful
   ---------------------------------------- */


BOOL EEPROMI2C_StoreData(UINT8* data, UINT16 startaddr,int ndata) {
    I2C_7_BIT_ADDRESS SlaveAddress;
    int i;
    int dataindex = 0;
    int nbytes;

    UINT8 hiaddr;
    UINT8 loaddr;


    // must write in block of 64 on addresses that are multiples of 64

    do {
        hiaddr = (UINT8) ((startaddr >> 8)& 0x00FF);
        loaddr = (UINT8) (startaddr & 0x00FF);

        nbytes = 64 - (loaddr & 0x003F);

        if ((dataindex + nbytes) > ndata) {
            nbytes = ndata - dataindex;
        }

        // can only push 64 bytes at a time
        // must stop on 64 byte borders = lower 5 bits

        I2C_FORMAT_7_BIT_ADDRESS(SlaveAddress, EEPROM_ADDRESS, I2C_WRITE);


        // Start the transfer to write data to the EEPROM
        if (!I2C_StartTransfer(FALSE))return FALSE;

        // transmit request
        I2C_TransmitOneByte(SlaveAddress.byte);
        if (!I2CByteWasAcknowledged(EEPROM_I2C_BUS)) return FALSE; // DBPRINTF("Error: Sent byte was not acknowledged\n");

        // transmit address
        I2C_TransmitOneByte(hiaddr);
        if (!I2CByteWasAcknowledged(EEPROM_I2C_BUS)) return FALSE; // DBPRINTF("Error: Sent byte was not acknowledged\n");

        I2C_TransmitOneByte(loaddr);
        if (!I2CByteWasAcknowledged(EEPROM_I2C_BUS)) return FALSE; // DBPRINTF("Error: Sent byte was not acknowledged\n");


        // Transmit all data
        for (i = 0; i < nbytes; i++) {
            // Transmit a byte
            if (!I2C_TransmitOneByte(data[dataindex])) return FALSE;
            // Verify that the byte was acknowledged
            if (!I2CByteWasAcknowledged(EEPROM_I2C_BUS)) return FALSE;
            // DBPRINTF("Error: Sent byte was not acknowledged\n");
            dataindex++;
        }

        // End the transfer (hang here if an error occured)
        I2C_StopTransfer();

        // Wait for EEPROM to complete write process, by polling the ack status.

        do {
            // Start the transfer to address the EEPROM
            if (!I2C_StartTransfer(FALSE)) return FALSE;

            // Transmit just the EEPROM's address
            if (!I2C_TransmitOneByte(SlaveAddress.byte)) return FALSE;
            // Check to see if the byte was acknowledged
            i = I2CByteWasAcknowledged(EEPROM_I2C_BUS);

            // End the transfer (stop here if an error occured)
            I2C_StopTransfer();
        } while (!i);


        // get ready for the next 64 bytes

        startaddr += nbytes;

    } while (dataindex < ndata);
}

/* --------------------------------------
 *
 *  Read array of data on EEPROM starting at specific address
 *
 *  Parameters:
 *     data = pointer to array of UNIT8.  Data will be returned in this array.
 *    addr =  address in eeprom to write data
 *     ndata = number of elements to write
 *
 *  Return: true if successful
   ---------------------------------------- */

BOOL EEPROMI2C_ReadData(UINT8* data,  UINT16 addr,int ndata) {
    I2C_7_BIT_ADDRESS SlaveAddress;
    int i;
    UINT8 hiaddr = (UINT8) (addr >> 8);
    UINT8 loaddr = (UINT8) (addr & 0x00FF);

    I2C_FORMAT_7_BIT_ADDRESS(SlaveAddress, EEPROM_ADDRESS, I2C_WRITE);


    // Start the transfer to read the EEPROM.
    if (!I2C_StartTransfer(FALSE)) return FALSE;

    // transmit request
    I2C_TransmitOneByte(SlaveAddress.byte);
    if (!I2CByteWasAcknowledged(EEPROM_I2C_BUS)) return FALSE; // DBPRINTF("Error: Sent byte was not acknowledged\n");

    // transmit address
    I2C_TransmitOneByte(hiaddr);
    if (!I2CByteWasAcknowledged(EEPROM_I2C_BUS)) return FALSE; // DBPRINTF("Error: Sent byte was not acknowledged\n");

    I2C_TransmitOneByte(loaddr);
    if (!I2CByteWasAcknowledged(EEPROM_I2C_BUS)) return FALSE; // DBPRINTF("Error: Sent byte was not acknowledged\n");


    if (!I2C_StartTransfer(TRUE)) return FALSE;
    // Transmit the address with the READ bit set
    I2C_FORMAT_7_BIT_ADDRESS(SlaveAddress, EEPROM_ADDRESS, I2C_READ);
    I2C_TransmitOneByte(SlaveAddress.byte);
    // Verify that the byte was acknowledged
    if (!I2CByteWasAcknowledged(EEPROM_I2C_BUS)) return FALSE;

    // get all but the last byte
    for (i = 0; i < (ndata - 1); i++) {
        if (I2CReceiverEnable(EEPROM_I2C_BUS, TRUE) == I2C_RECEIVE_OVERFLOW) return FALSE;
        // DBPRINTF("Error: I2C Receive Overflow\n");


        while (!I2CReceivedDataIsAvailable(EEPROM_I2C_BUS));

        I2CAcknowledgeByte(EEPROM_I2C_BUS, TRUE);
        data[i] = I2CGetByte(EEPROM_I2C_BUS);

        while (!I2CAcknowledgeHasCompleted(EEPROM_I2C_BUS));


    }
    if (I2CReceiverEnable(EEPROM_I2C_BUS, TRUE) == I2C_RECEIVE_OVERFLOW) return FALSE;

    // get the last byte and stop
    while (!I2CReceivedDataIsAvailable(EEPROM_I2C_BUS));
    I2CAcknowledgeByte(EEPROM_I2C_BUS, FALSE);
    data[ndata - 1] = I2CGetByte(EEPROM_I2C_BUS);
    while (!I2CAcknowledgeHasCompleted(EEPROM_I2C_BUS));


    // End the transfer (stop here if an error occured)
    I2C_StopTransfer();

}



/* --------------------------------------
 *
 *  Set the DAC output.  Don't store in the EEPROM
 *  This is set for "fast mode" - so doesn't store on the device EEPROM
 *  First call set startStream to true.  Successive calls set startStream to false
 *   must call I2C_StopTransfer when done.  Don't access any other i2c device until
 *  I2C_StopTransfer has been sent.
 *
 *  Parameters:
 *     data =       value to set
 *      startStream = send the device address before sending data.
 *
 *  Return: true if successful
   ---------------------------------------- */


BOOL DACI2C_SetValue(int data, BOOL startStream) {
    I2C_7_BIT_ADDRESS SlaveAddress;


    UINT8 hibyte;
    UINT8 lobyte;



    hibyte = (UINT8) ((data >> 8)& 0x000F); // first three bits are mode bits
                                            // these can be changed if you also want to store the value to the EEPROM
                                            // on the DAC.  Nominal values are all zero.
                                            // see the Microchip MPC4725 DAC documentation for details
    lobyte = (UINT8) (data & 0x00FF);

    if (startStream) {
        I2C_FORMAT_7_BIT_ADDRESS(SlaveAddress, DAC_ADDRESS, I2C_WRITE);


        // Start the transfer to write data to the DAC
        if (!I2C_StartTransfer(FALSE))return FALSE;


        // transmit address
        I2C_TransmitOneByte(SlaveAddress.byte);
        if (!I2CByteWasAcknowledged(EEPROM_I2C_BUS)) return FALSE; // DBPRINTF("Error: Sent byte was not acknowledged\n");
    }
    // transmit data
    I2C_TransmitOneByte(hibyte);
    if (!I2CByteWasAcknowledged(EEPROM_I2C_BUS)) return FALSE; // DBPRINTF("Error: Sent byte was not acknowledged\n");

    I2C_TransmitOneByte(lobyte);
    if (!I2CByteWasAcknowledged(EEPROM_I2C_BUS)) return FALSE; // DBPRINTF("Error: Sent byte was not acknowledged\n");

 //   if (endStream) {
        // End the transfer (hang here if an error occured)
//        DACI2C_StopTransfer();
//    }

}
