/* -----------------------------------------------------------
 * 
 *  main.c -- Uses MX4cK as flash drive.
 *  Lets you store data on the EEPROM and view or edit it on a PC over
 *  a USB connection.
 *
 * IMPORTANT: MX4cK JP6 must be in OTG position
 * 
 *  This example uses two threads, one to monitor and handle the Mass Storage Device
 *  the other flashes a LED - a timer. USB handing is done through an ISR.  The ISR has been
 *  wrapped to be compatible with the RTOS. The USB and MSD functions are blocking
 *
 * In this implementation the USB interrupt runs the USB code directly.  An alternate method is
 * to use the interrupt to signal another task through a semaphore, then handle the USB action inside
 * of the task.  That approach is the preferred why of handling interrupts in the ROTS. See the
 * Interrupt example.
 *
 *  Data is stored on the EEPROM as a single flat file.  When connected to a PC, the MX4cK
 *  reports being a FAT12 device. See the Readme.txt in USBDevice for details. 
 * 
 * Based on FreeRTOS V7.3.0 and the Microchip USB library
 * See individual files for Copyright notices
 *
 * BUG POSSIBILITY ALERT:  Internal Flash.c  ReadSector function has not been tested for files
 * larger than 512 bytes.
 * 
 * Mason, Seattle U. 2013
 * 
 * ----------------------------------------------------------- */


#include "MX4cKsetup.h"
#include "MX4cKsettings.h"

#include "RTOS/FreeRTOS.h"
#include "RTOS/task.h"
#include "RTOS/queue.h"
#include "RTOS/timers.h"


#include "USBDevice/HardwareProfile.h"
#include "USBDevice/usb.h"


// PRIVATE PROTOTYPES 
static void InitializeSystem(void);
void ProcessIO(void);
static void myTimer(xTimerHandle xTimer);
static void myTask(void *pvParameters);
static void usbTask(void *pvParameters);

//-----------------------------------------------------------
int main(void) {
    xTimerHandle xTimer1;


     // Configure built-in LEDs
    //   LED1 is on RB10
    //	 LED2 is on RB11
    //	 LED3 is on RB12
    //	 LED4 is on RB13
    //Set ports for onboard LEDs to outputs
    PORTSetPinsDigitalOut(IOPORT_B, BIT_10 | BIT_11 | BIT_12 | BIT_13);
    PORTClearBits(IOPORT_B, BIT_10 | BIT_11 | BIT_12 | BIT_13);

    InitializeSystem();

    // set up a timer to flash the LED
    xTimer1 = xTimerCreate((const signed char *) "yyTimer", // A text name, purely to help debugging.
            (500), // The timer period in tick counts  (5000 * 1ms = 1sec)
            pdTRUE, // This is an auto-reload timer, so xAutoReload is set to pdTRUE.
            (void *) 0, // The ID is not used, so can be set to anything.
            myTimer // The callback function that inspects the status of all the other tasks.
            );

    // make sure the timer was created the start it
    if (xTimer1 != NULL) {
        xTimerStart(xTimer1, 0);
    }

    // create a thread to handle the mass storage device requests
    xTaskCreate(usbTask, // The function that implements the task.
            (signed char *) "usbTask", // The text name assigned to the task - for debug only as it is not used by the kernel.
            configMINIMAL_STACK_SIZE+250, // Allocate extra space for USB functions 
            (void *) (0), // The parameter passed to the task - just to check the functionality.
            1, // The priority assigned to the task. Currently lower than the timer
            NULL); // The task handle is not required, so NULL is passed.


    // initialize I2C bus to the EEPROM
    I2C_Initialize();
    // could read and write from the EEPROM in the program too
    // using  EEPROMI2C_StoreData and EEPROMI2C_ReadData
    // See I2C_EEPROM files and USBDevice readme.txt for explanation of how
    // data is organized on the EEPROM in this example
    // When accessing the EEPROM in a thread/timer you will need to use a mutex to prevent the 
    // MSD and USB functions from accessing the I2C at the same time.
   

    // initialize the USB OTG and start the USB interrupt
    // USB ISR is   _USB1Interrupt in usb_device.c
    // is it wrapped inside a RTOS compatible ISR  usb_isr.S
    USBDeviceInit();
    USBDeviceAttach();

    /* Start the tasks and timer running. */
    vTaskStartScheduler();


    for (;;); // never get here
    return 0;

}//end main

/* -----------------------------------------------------------
 * Function:        static void InitializeSystem(void)
 *
 * Overview:        InitializeSystem is a centralize initialization
 *                  routine. All required USB initialization routines
 *                  are called from here.
 *
 *                  User application initialization routine should
 *                  also be called from here.                  
 ----------------------------------------------------------- */
static void InitializeSystem(void) {
    SYSTEMConfigPerformance(SYSCLOCK);
    mOSCSetPBDIV(OSC_PB_DIV_8); // the peripherial clock is 10MHz = 80MHz/8

    /* Setup to use the external interrupt controller. */
    INTEnableSystemMultiVectoredInt();

    // portDISABLE_INTERRUPTS();
}

/* -----------------------------------------------------------
 * Monitor the usb state and handle disk tasks as needed
 * This task blocks other tasks when handling the USB MSD
 ----------------------------------------------------------- */
static void usbTask(void *pvParameters) {
    for (;;) {
        // check to see if the MSD needs servicing
        if (!((USBDeviceState < CONFIGURED_STATE) || (USBSuspendControl == 1))) {         
            taskENTER_CRITICAL(); // prevent context switches during MSD handling
            MSDTasks();
            taskEXIT_CRITICAL();           
        } else {
            // nothing to do now, so give up control of the CPU to other tasks
            taskYIELD();         
        }
    }
}

/* -----------------------------------------------------------
 *  Flash the LED
 *  LED flashing will be irregular when USB or MSB are very active
 * ----------------------------------------------------------- */
static void myTimer(xTimerHandle xTimer) {
    // This function is called everytime the software time expires.
    // timers are set as priority 2
    PORTToggleBits(IOPORT_B, BIT_10);
}

/*-----------------------------------------------------------
 *  Hooks for RTOS.  Are not required to modify these
 *  May want to turn on LED or some other external signal if an error
 *  occurs
 *-----------------------------------------------------------*/

void vApplicationMallocFailedHook(void) {
    // support function.  See RTOS Starter
    // if an error occurs go into an infinite loop
    taskDISABLE_INTERRUPTS();
    for (;;);
}

void vApplicationIdleHook(void) {
    // support function.  See RTOS Starter
}

void vApplicationStackOverflowHook(xTaskHandle pxTask, signed char *pcTaskName) {
    (void) pcTaskName;
    (void) pxTask;
    // support function.  See RTOS Starter
    // if an error occurs go into an infinite loop
    taskDISABLE_INTERRUPTS();
    for (;;);
}

void vApplicationTickHook(void) {
    // support function.  See RTOS Starter
}

void _general_exception_handler(unsigned long ulCause, unsigned long ulStatus) {
    // support function.  See RTOS Starter
    // if an error occurs go into an infinite loop
    for (;;);
}


/** EOF main.c ***************************************************************/

